@@ -1,33 +1,47 @@
 /*
 	frame: Heap of routines dealing with the core mpg123 data structure.
 
-	copyright 2008-9 by the mpg123 project - free software under the terms of the LGPL 2.1
+	copyright 2008-2023 by the mpg123 project - free software under the terms of the LGPL 2.1
 	see COPYING and AUTHORS files in distribution or http://mpg123.org
 	initially written by Thomas Orgis
 */
 
+#define WANT_GETCPUFLAGS
 #include "mpg123lib_intern.h"
 #include "getcpuflags.h"
 #include "debug.h"
 
-#define IGNORESHIFT 2
-
 static void frame_fixed_reset(mpg123_handle *fr);
 
 /* that's doubled in decode_ntom.c */
 #define NTOM_MUL (32768)
-#define aligned_pointer(p,type,alignment) \
-	(((char*)(p)-(char*)NULL) % (alignment)) \
-	? (type*)((char*)(p) + (alignment) - (((char*)(p)-(char*)NULL) % (alignment))) \
-	: (type*)(p)
-void frame_default_pars(mpg123_pars *mp)
+
+#define aligned_pointer(p, type, alignment) align_the_pointer(p, alignment)
+static void *align_the_pointer(void *base, unsigned int alignment)
+{
+	/*
+		Work in unsigned integer realm, explicitly.
+		Tricking the compiler into integer operations like % by invoking base-NULL is dangerous: It results into ptrdiff_t, which gets negative on big addresses. Big screw up, that.
+		I try to do it "properly" here: Casting only to uintptr_t and no artihmethic with void*.
+	*/
+	uintptr_t baseval = (uintptr_t)(char*)base;
+	uintptr_t aoff = baseval % alignment;
+
+	debug3("align_the_pointer: pointer %p is off by %u from %u",
+	       base, (unsigned int)aoff, alignment);
+
+	if(aoff) return (char*)base+alignment-aoff;
+	else     return base;
+}
+
+static void frame_default_pars(mpg123_pars *mp)
 {
 	mp->outscale = 1.0;
-#ifdef GAPLESS
-	mp->flags = MPG123_GAPLESS;
-#else
 	mp->flags = 0;
+#ifdef GAPLESS
+	mp->flags |= MPG123_GAPLESS;
 #endif
+	mp->flags |= MPG123_AUTO_RESAMPLE|MPG123_FLOAT_FALLBACK;
 #ifndef NO_NTOM
 	mp->force_rate = 0;
 #endif
@@ -39,71 +53,113 @@ void frame_default_pars(mpg123_pars *mp)
 #ifndef NO_ICY
 	mp->icy_interval = 0;
 #endif
-#ifndef WIN32
 	mp->timeout = 0;
-#endif
 	mp->resync_limit = 1024;
 #ifdef FRAME_INDEX
 	mp->index_size = INDEX_SIZE;
 #endif
+	mp->preframes = 4; /* That's good  for layer 3 ISO compliance bitstream. */
 	mpg123_fmt_all(mp);
+	/* Default of keeping some 4K buffers at hand, should cover the "usual" use case (using 16K pipe buffers as role model). */
+#ifndef NO_FEEDER
+	mp->feedpool = 5; 
+	mp->feedbuffer = 4096;
+#endif
+	mp->freeformat_framesize = -1;
 }
 
-void frame_init(mpg123_handle *fr)
+void INT123_frame_init(mpg123_handle *fr)
 {
-	frame_init_par(fr, NULL);
+	INT123_frame_init_par(fr, NULL);
 }
 
-void frame_init_par(mpg123_handle *fr, mpg123_pars *mp)
+void INT123_frame_init_par(mpg123_handle *fr, mpg123_pars *mp)
 {
-	fr->own_buffer = FALSE;
+	fr->own_buffer = TRUE;
 	fr->buffer.data = NULL;
+	fr->buffer.rdata = NULL;
+	fr->buffer.fill = 0;
+	fr->buffer.size = 0;
 	fr->rawbuffs = NULL;
 	fr->rawbuffss = 0;
 	fr->rawdecwin = NULL;
 	fr->rawdecwins = 0;
+#ifdef REAL_IS_FIXED
+	fr->gainpow2 = NULL; // At least crash early if I get it wrong.
+#endif
 #ifndef NO_8BIT
 	fr->conv16to8_buf = NULL;
 #endif
+#ifdef OPT_DITHER
+	fr->dithernoise = NULL;
+#endif
+	fr->layerscratch = NULL;
 	fr->xing_toc = NULL;
-	fr->cpu_opts.type = defdec();
-	fr->cpu_opts.class = decclass(fr->cpu_opts.type);
+#ifdef OPT_CPU_FLAGS
+	wrap_getcpuflags(&(fr->cpu_flags));
+#endif
+	fr->cpu_opts.type = INT123_defdec();
+	fr->cpu_opts.class = INT123_decclass(fr->cpu_opts.type);
 #ifndef NO_NTOM
-	/* these two look unnecessary, check guarantee for synth_ntom_set_step (in control_generic, even)! */
-	fr->ntom_val[0] = NTOM_MUL>>1;
-	fr->ntom_val[1] = NTOM_MUL>>1;
+	/* these two look unnecessary, check guarantee for INT123_synth_ntom_set_step (in control_generic, even)! */
+	fr->INT123_ntom_val[0] = NTOM_MUL>>1;
+	fr->INT123_ntom_val[1] = NTOM_MUL>>1;
 	fr->ntom_step = NTOM_MUL;
 #endif
 	/* unnecessary: fr->buffer.size = fr->buffer.fill = 0; */
 	mpg123_reset_eq(fr);
-	init_icy(&fr->icy);
-	init_id3(fr);
-	/* frame_outbuffer is missing... */
-	/* frame_buffers is missing... that one needs cpu opt setting! */
-	/* after these... frame_reset is needed before starting full decode */
-	invalidate_format(&fr->af);
-	fr->rdat.r_read = NULL;
-	fr->rdat.r_lseek = NULL;
+	INT123_init_icy(&fr->icy);
+	INT123_init_id3(fr);
+	/* INT123_frame_outbuffer is missing... */
+	/* INT123_frame_buffers is missing... that one needs cpu opt setting! */
+	/* after these... INT123_frame_reset is needed before starting full decode */
+	INT123_invalidate_format(&fr->af);
+	fr->rdat.iohandle = NULL;
+	fr->rdat.r_read64 = NULL;
+	fr->rdat.r_lseek64 = NULL;
+	fr->rdat.cleanup_handle = NULL;
+	fr->wrapperdata = NULL;
 	fr->decoder_change = 1;
 	fr->err = MPG123_OK;
 	if(mp == NULL) frame_default_pars(&fr->p);
 	else memcpy(&fr->p, mp, sizeof(struct mpg123_pars_struct));
 
+#ifndef NO_FEEDER
+	INT123_bc_prepare(&fr->rdat.buffer, fr->p.feedpool, fr->p.feedbuffer);
+#endif
+
 	fr->down_sample = 0; /* Initialize to silence harmless errors when debugging. */
+	fr->id3v2_raw = NULL;
 	frame_fixed_reset(fr); /* Reset only the fixed data, dynamic buffers are not there yet! */
 	fr->synth = NULL;
 	fr->synth_mono = NULL;
-	fr->make_decode_tables = NULL;
+	fr->INT123_make_decode_tables = NULL;
 #ifdef FRAME_INDEX
-	fi_init(&fr->index);
-	frame_index_setup(fr); /* Apply the size setting. */
+	INT123_fi_init(&fr->index);
+	INT123_frame_index_setup(fr); /* Apply the size setting. */
 #endif
-#ifdef OPT_DITHER
-	/* The idea is to read that at runtime in the future... to avoid bloating the binary. */
-	fr->dithernoise = dithernoise;
+#ifndef NO_MOREINFO
+	fr->pinfo = NULL;
 #endif
 }
 
+#ifdef OPT_DITHER
+/* Also, only allocate the memory for the table on demand.
+   In future, one could create special noise for different sampling frequencies(?). */
+int INT123_frame_dither_init(mpg123_handle *fr)
+{
+	/* run-time dither noise table generation */
+	if(fr->dithernoise == NULL)
+	{
+		fr->dithernoise = malloc(sizeof(float)*DITHERSIZE);
+		if(fr->dithernoise == NULL) return 0;
+
+		INT123_dither_table_init(fr->dithernoise);
+	}
+	return 1;
+}
+#endif
+
 mpg123_pars attribute_align_arg *mpg123_new_pars(int *error)
 {
 	mpg123_pars *mp = malloc(sizeof(struct mpg123_pars_struct));
@@ -119,43 +175,64 @@ void attribute_align_arg mpg123_delete_pars(mpg123_pars* mp)
 
 int attribute_align_arg mpg123_reset_eq(mpg123_handle *mh)
 {
+	if(mh == NULL) return MPG123_BAD_HANDLE;
+#ifndef NO_EQUALIZER
 	int i;
 	mh->have_eq_settings = 0;
 	for(i=0; i < 32; ++i) mh->equalizer[0][i] = mh->equalizer[1][i] = DOUBLE_TO_REAL(1.0);
-
+#endif
 	return MPG123_OK;
 }
 
-int frame_outbuffer(mpg123_handle *fr)
+int INT123_frame_outbuffer(mpg123_handle *fr)
 {
-	size_t size = mpg123_safe_buffer()*AUDIOBUFSIZE;
-	if(!fr->own_buffer) fr->buffer.data = NULL;
-	if(fr->buffer.data != NULL && fr->buffer.size != size)
+	size_t size = fr->outblock;
+	if(!fr->own_buffer)
 	{
-		free(fr->buffer.data);
-		fr->buffer.data = NULL;
+		if(fr->buffer.size < size)
+		{
+			fr->err = MPG123_BAD_BUFFER;
+			if(NOQUIET)
+				merror( "have external buffer of size %"SIZE_P", need %"SIZE_P
+				,	(size_p)fr->buffer.size, (size_p)size );
+			return MPG123_ERR;
+		}
+	}
+
+	debug1("need frame buffer of %"SIZE_P, (size_p)size);
+	if(fr->buffer.rdata != NULL && fr->buffer.size != size)
+	{
+		free(fr->buffer.rdata);
+		fr->buffer.rdata = NULL;
 	}
 	fr->buffer.size = size;
-	if(fr->buffer.data == NULL) fr->buffer.data = (unsigned char*) malloc(fr->buffer.size);
-	if(fr->buffer.data == NULL)
+	fr->buffer.data = NULL;
+	/* be generous: use 16 byte alignment */
+	if(fr->buffer.rdata == NULL) fr->buffer.rdata = (unsigned char*) malloc(fr->buffer.size+15);
+	if(fr->buffer.rdata == NULL)
 	{
 		fr->err = MPG123_OUT_OF_MEM;
-		return -1;
+		return MPG123_ERR;
 	}
+	fr->buffer.data = aligned_pointer(fr->buffer.rdata, unsigned char*, 16);
 	fr->own_buffer = TRUE;
 	fr->buffer.fill = 0;
-	return 0;
+	return MPG123_OK;
 }
 
-int attribute_align_arg mpg123_replace_buffer(mpg123_handle *mh, unsigned char *data, size_t size)
+int attribute_align_arg mpg123_replace_buffer(mpg123_handle *mh, void *data, size_t size)
 {
-	if(data == NULL || size < mpg123_safe_buffer())
+	debug2("replace buffer with %p size %"SIZE_P, data, (size_p)size);
+	if(mh == NULL) return MPG123_BAD_HANDLE;
+	/* Will accept any size, the error comes later... */
+	if(data == NULL)
 	{
 		mh->err = MPG123_BAD_BUFFER;
 		return MPG123_ERR;
 	}
-	if(mh->own_buffer && mh->buffer.data != NULL) free(mh->buffer.data);
+	if(mh->buffer.rdata != NULL) free(mh->buffer.rdata);
 	mh->own_buffer = FALSE;
+	mh->buffer.rdata = NULL;
 	mh->buffer.data = data;
 	mh->buffer.size = size;
 	mh->buffer.fill = 0;
@@ -163,36 +240,36 @@ int attribute_align_arg mpg123_replace_buffer(mpg123_handle *mh, unsigned char *
 }
 
 #ifdef FRAME_INDEX
-int frame_index_setup(mpg123_handle *fr)
+int INT123_frame_index_setup(mpg123_handle *fr)
 {
 	int ret = MPG123_ERR;
 	if(fr->p.index_size >= 0)
 	{ /* Simple fixed index. */
 		fr->index.grow_size = 0;
-		debug1("resizing index to %li", fr->p.index_size);
-		ret = fi_resize(&fr->index, (size_t)fr->p.index_size);
-		debug2("index resized... %lu at %p", (unsigned long)fr->index.size, (void*)fr->index.data);
+		ret = INT123_fi_resize(&fr->index, (size_t)fr->p.index_size);
 	}
 	else
 	{ /* A growing index. We give it a start, though. */
 		fr->index.grow_size = (size_t)(- fr->p.index_size);
 		if(fr->index.size < fr->index.grow_size)
-		ret = fi_resize(&fr->index, fr->index.grow_size);
+			ret = INT123_fi_resize(&fr->index, fr->index.grow_size);
 		else
-		ret = MPG123_OK; /* We have minimal size already... and since growing is OK... */
+			ret = MPG123_OK; /* We have minimal size already... and since growing is OK... */
 	}
 	debug2("set up frame index of size %lu (ret=%i)", (unsigned long)fr->index.size, ret);
-
+	if(ret && NOQUIET)
+		error("frame index setup (initial resize) failed");
 	return ret;
 }
 #endif
 
 static void frame_decode_buffers_reset(mpg123_handle *fr)
 {
-	memset(fr->rawbuffs, 0, fr->rawbuffss);
+	if(fr->rawbuffs) /* memset(NULL, 0, 0) not desired */
+		memset(fr->rawbuffs, 0, fr->rawbuffss);
 }
 
-int frame_buffers(mpg123_handle *fr)
+int INT123_frame_buffers(mpg123_handle *fr)
 {
 	int buffssize = 0;
 	debug1("frame %p buffer", (void*)fr);
@@ -268,11 +345,19 @@ int frame_buffers(mpg123_handle *fr)
 #endif
 			/* decwin_mmx will share, decwins will be appended ... sizeof(float)==4 */
 			if(decwin_size < (512+32)*4) decwin_size = (512+32)*4;
-			decwin_size += (512+32)*4 + 31; /* the second window + alignment zone */
-			/* (512+32)*4/32 == 2176/32 == 68, so one decwin block retains alignment */
+
+			/* the second window + alignment zone -- we align for 32 bytes for SSE as
+			   requirement, 64 byte for matching cache line size (that matters!) */
+			decwin_size += (512+32)*4 + 63;
+			/* (512+32)*4/32 == 2176/32 == 68, so one decwin block retains alignment for 32 or 64 bytes */
 #ifdef OPT_MULTI
 		}
 #endif
+#endif
+#if defined(OPT_ALTIVEC) || defined(OPT_ARM) 
+		/* sizeof(real) >= 4 ... yes, it could be 8, for example.
+		   We got it intialized to at least (512+32)*sizeof(real).*/
+		decwin_size += 512*sizeof(real);
 #endif
 		/* Hm, that's basically realloc() ... */
 		if(fr->rawdecwin != NULL && fr->rawdecwins != decwin_size)
@@ -294,8 +379,8 @@ int frame_buffers(mpg123_handle *fr)
 		{
 #endif
 			/* align decwin, assign that to decwin_mmx, append decwins */
-			/* I need to add to decwin what is missing to the next full 32 byte -- also I want to make gcc -pedantic happy... */
-			fr->decwin = aligned_pointer(fr->rawdecwin,real,32);
+			/* I need to add to decwin what is missing to the next full 64 byte -- also I want to make gcc -pedantic happy... */
+			fr->decwin = aligned_pointer(fr->rawdecwin,real,64);
 			debug1("aligned decwin: %p", (void*)fr->decwin);
 			fr->decwin_mmx = (float*)fr->decwin;
 			fr->decwins = fr->decwin_mmx+512+32;
@@ -305,6 +390,51 @@ int frame_buffers(mpg123_handle *fr)
 #endif
 #endif
 	}
+
+	/* Layer scratch buffers are of compile-time fixed size, so allocate only once. */
+	if(fr->layerscratch == NULL)
+	{
+		/* Allocate specific layer1/2/3 buffers, so that we know they'll work for SSE. */
+		size_t scratchsize = 0;
+		real *scratcher;
+#ifndef NO_LAYER1
+		scratchsize += sizeof(real) * 2 * SBLIMIT;
+#endif
+#ifndef NO_LAYER2
+		scratchsize += sizeof(real) * 2 * 4 * SBLIMIT;
+#endif
+#ifndef NO_LAYER3
+		scratchsize += sizeof(real) * 2 * SBLIMIT * SSLIMIT; /* hybrid_in */
+		scratchsize += sizeof(real) * 2 * SSLIMIT * SBLIMIT; /* hybrid_out */
+#endif
+		/*
+			Now figure out correct alignment:
+			We need 16 byte minimum, smallest unit of the blocks is 2*SBLIMIT*sizeof(real), which is 64*4=256. Let's do 64bytes as heuristic for cache line (as proven useful in buffs above).
+		*/
+		fr->layerscratch = malloc(scratchsize+63);
+		if(fr->layerscratch == NULL) return -1;
+
+		/* Get aligned part of the memory, then divide it up. */
+		scratcher = aligned_pointer(fr->layerscratch,real,64);
+		/* Those funky pointer casts silence compilers...
+		   One might change the code at hand to really just use 1D arrays, but in practice, that would not make a (positive) difference. */
+#ifndef NO_LAYER1
+		fr->layer1.fraction = (real(*)[SBLIMIT])scratcher;
+		scratcher += 2 * SBLIMIT;
+#endif
+#ifndef NO_LAYER2
+		fr->layer2.fraction = (real(*)[4][SBLIMIT])scratcher;
+		scratcher += 2 * 4 * SBLIMIT;
+#endif
+#ifndef NO_LAYER3
+		fr->layer3.hybrid_in = (real(*)[SBLIMIT][SSLIMIT])scratcher;
+		scratcher += 2 * SBLIMIT * SSLIMIT;
+		fr->layer3.hybrid_out = (real(*)[SSLIMIT][SBLIMIT])scratcher;
+		scratcher += 2 * SSLIMIT * SBLIMIT;
+#endif
+		/* Note: These buffers don't need resetting here. */
+	}
+
 	/* Only reset the buffers we created just now. */
 	frame_decode_buffers_reset(fr);
 
@@ -312,14 +442,14 @@ int frame_buffers(mpg123_handle *fr)
 	return 0;
 }
 
-int frame_buffers_reset(mpg123_handle *fr)
+int INT123_frame_buffers_reset(mpg123_handle *fr)
 {
 	fr->buffer.fill = 0; /* hm, reset buffer fill... did we do a flush? */
 	fr->bsnum = 0;
 	/* Wondering: could it be actually _wanted_ to retain buffer contents over different files? (special gapless / cut stuff) */
 	fr->bsbuf = fr->bsspace[1];
 	fr->bsbufold = fr->bsbuf;
-	fr->bitreservoir = 0; /* Not entirely sure if this is the right place for that counter. */
+	fr->bitreservoir = 0;
 	frame_decode_buffers_reset(fr);
 	memset(fr->bsspace, 0, 2*(MAXFRAMESIZE+512));
 	memset(fr->ssave, 0, 34);
@@ -328,7 +458,7 @@ int frame_buffers_reset(mpg123_handle *fr)
 	return 0;
 }
 
-void frame_icy_reset(mpg123_handle* fr)
+static void frame_icy_reset(mpg123_handle* fr)
 {
 #ifndef NO_ICY
 	if(fr->icy.data != NULL) free(fr->icy.data);
@@ -338,13 +468,13 @@ void frame_icy_reset(mpg123_handle* fr)
 #endif
 }
 
-void frame_free_toc(mpg123_handle *fr)
+static void frame_free_toc(mpg123_handle *fr)
 {
 	if(fr->xing_toc != NULL){ free(fr->xing_toc); fr->xing_toc = NULL; }
 }
 
 /* Just copy the Xing TOC over... */
-int frame_fill_toc(mpg123_handle *fr, unsigned char* in)
+int INT123_frame_fill_toc(mpg123_handle *fr, unsigned char* in)
 {
 	if(fr->xing_toc == NULL) fr->xing_toc = malloc(100);
 	if(fr->xing_toc != NULL)
@@ -365,13 +495,13 @@ int frame_fill_toc(mpg123_handle *fr, unsigned char* in)
 
 /* Prepare the handle for a new track.
    Reset variables, buffers... */
-int frame_reset(mpg123_handle* fr)
+int INT123_frame_reset(mpg123_handle* fr)
 {
-	frame_buffers_reset(fr);
+	INT123_frame_buffers_reset(fr);
 	frame_fixed_reset(fr);
 	frame_free_toc(fr);
 #ifdef FRAME_INDEX
-	fi_reset(&fr->index);
+	INT123_fi_reset(&fr->index);
 #endif
 
 	return 0;
@@ -381,19 +511,21 @@ int frame_reset(mpg123_handle* fr)
 static void frame_fixed_reset(mpg123_handle *fr)
 {
 	frame_icy_reset(fr);
-	open_bad(fr);
+	INT123_open_bad(fr);
 	fr->to_decode = FALSE;
 	fr->to_ignore = FALSE;
 	fr->metaflags = 0;
-	fr->outblock = mpg123_safe_buffer();
+	fr->outblock = 0; /* This will be set before decoding! */
 	fr->num = -1;
+	fr->input_offset = -1;
 	fr->playnum = -1;
-	fr->accurate = TRUE;
+	fr->state_flags = FRAME_ACCURATE;
 	fr->silent_resync = 0;
 	fr->audio_start = 0;
 	fr->clip = 0;
 	fr->oldhead = 0;
 	fr->firsthead = 0;
+	fr->lay = 0;
 	fr->vbr = MPG123_CBR;
 	fr->abr_rate = 0;
 	fr->track_frames = 0;
@@ -411,12 +543,13 @@ static void frame_fixed_reset(mpg123_handle *fr)
 	fr->rva.peak[1] = 0;
 	fr->fsizeold = 0;
 	fr->firstframe = 0;
-	fr->ignoreframe = fr->firstframe-IGNORESHIFT;
+	fr->ignoreframe = fr->firstframe-fr->p.preframes;
+	fr->header_change = 0;
 	fr->lastframe = -1;
 	fr->fresh = 1;
 	fr->new_format = 0;
 #ifdef GAPLESS
-	frame_gapless_init(fr,0,0);
+	INT123_frame_gapless_init(fr,-1,0,0);
 	fr->lastoff = 0;
 	fr->firstoff = 0;
 #endif
@@ -427,8 +560,8 @@ static void frame_fixed_reset(mpg123_handle *fr)
 #ifdef OPT_DITHER
 	fr->ditherindex = 0;
 #endif
-	reset_id3(fr);
-	reset_icy(&fr->icy);
+	INT123_reset_id3(fr);
+	INT123_reset_icy(&fr->icy);
 	/* ICY stuff should go into icy.c, eh? */
 #ifndef NO_ICY
 	fr->icy.interval = 0;
@@ -436,9 +569,17 @@ static void frame_fixed_reset(mpg123_handle *fr)
 #endif
 	fr->halfphase = 0; /* here or indeed only on first-time init? */
 	fr->error_protection = 0;
+	fr->freeformat_framesize = fr->p.freeformat_framesize;
+	fr->enc_delay = -1;
+	fr->enc_padding = -1;
+	memset(fr->id3buf, 0, sizeof(fr->id3buf));
+	if(fr->id3v2_raw)
+		free(fr->id3v2_raw);
+	fr->id3v2_raw = NULL;
+	fr->id3v2_size = 0;
 }
 
-void frame_free_buffers(mpg123_handle *fr)
+static void frame_free_buffers(mpg123_handle *fr)
 {
 	if(fr->rawbuffs != NULL) free(fr->rawbuffs);
 	fr->rawbuffs = NULL;
@@ -450,58 +591,59 @@ void frame_free_buffers(mpg123_handle *fr)
 	if(fr->conv16to8_buf != NULL) free(fr->conv16to8_buf);
 	fr->conv16to8_buf = NULL;
 #endif
+	if(fr->layerscratch != NULL) free(fr->layerscratch);
 }
 
-void frame_exit(mpg123_handle *fr)
+void INT123_frame_exit(mpg123_handle *fr)
 {
-	if(fr->own_buffer && fr->buffer.data != NULL)
+	if(fr->buffer.rdata != NULL)
 	{
-		debug1("freeing buffer at %p", (void*)fr->buffer.data);
-		free(fr->buffer.data);
+		debug1("freeing buffer at %p", (void*)fr->buffer.rdata);
+		free(fr->buffer.rdata);
 	}
-	fr->buffer.data = NULL;
+	fr->buffer.rdata = NULL;
 	frame_free_buffers(fr);
 	frame_free_toc(fr);
 #ifdef FRAME_INDEX
-	fi_exit(&fr->index);
+	INT123_fi_exit(&fr->index);
+#endif
+#ifdef OPT_DITHER
+	if(fr->dithernoise != NULL)
+	{
+		free(fr->dithernoise);
+		fr->dithernoise = NULL;
+	}
+#endif
+	INT123_exit_id3(fr);
+	INT123_clear_icy(&fr->icy);
+#ifndef NO_FEEDER
+	INT123_bc_cleanup(&fr->rdat.buffer);
 #endif
-	exit_id3(fr);
-	clear_icy(&fr->icy);
 }
 
-int attribute_align_arg mpg123_info(mpg123_handle *mh, struct mpg123_frameinfo *mi)
+int attribute_align_arg mpg123_framedata(mpg123_handle *mh, unsigned long *header, unsigned char **bodydata, size_t *bodybytes)
 {
-	if(mh == NULL) return MPG123_ERR;
-	if(mi == NULL)
-	{
-		mh->err = MPG123_ERR_NULL;
-		return MPG123_ERR;
-	}
-	mi->version = mh->mpeg25 ? MPG123_2_5 : (mh->lsf ? MPG123_2_0 : MPG123_1_0);
-	mi->layer = mh->lay;
-	mi->rate = frame_freq(mh);
-	switch(mh->mode)
-	{
-		case 0: mi->mode = MPG123_M_STEREO; break;
-		case 1: mi->mode = MPG123_M_JOINT;  break;
-		case 2: mi->mode = MPG123_M_DUAL;   break;
-		case 3: mi->mode = MPG123_M_MONO;   break;
-		default: error("That mode cannot be!");
-	}
-	mi->mode_ext = mh->mode_ext;
-	mi->framesize = mh->framesize+4; /* Include header. */
-	mi->flags = 0;
-	if(mh->error_protection) mi->flags |= MPG123_CRC;
-	if(mh->copyright)        mi->flags |= MPG123_COPYRIGHT;
-	if(mh->extension)        mi->flags |= MPG123_PRIVATE;
-	if(mh->original)         mi->flags |= MPG123_ORIGINAL;
-	mi->emphasis = mh->emphasis;
-	mi->bitrate  = frame_bitrate(mh);
-	mi->abr_rate = mh->abr_rate;
-	mi->vbr = mh->vbr;
+	if(mh == NULL)     return MPG123_BAD_HANDLE;
+	if(!mh->to_decode) return MPG123_ERR;
+
+	if(header    != NULL) *header    = mh->oldhead;
+	if(bodydata  != NULL) *bodydata  = mh->bsbuf;
+	if(bodybytes != NULL) *bodybytes = mh->framesize;
+
 	return MPG123_OK;
 }
 
+int attribute_align_arg mpg123_set_moreinfo( mpg123_handle *mh
+,	struct mpg123_moreinfo *mi)
+{
+#ifndef NO_MOREINFO
+	mh->pinfo = mi;
+	return MPG123_OK;
+#else
+	mh->err = MPG123_MISSING_FEATURE;
+	return MPG123_ERR;
+#endif
+}
 
 /*
 	Fuzzy frame offset searching (guessing).
@@ -511,10 +653,10 @@ int attribute_align_arg mpg123_info(mpg123_handle *mh, struct mpg123_frameinfo *
 		- guess wildly from mean framesize and offset of first frame / beginning of file.
 */
 
-off_t frame_fuzzy_find(mpg123_handle *fr, off_t want_frame, off_t* get_frame)
+static int64_t frame_fuzzy_find(mpg123_handle *fr, int64_t want_frame, int64_t* get_frame)
 {
 	/* Default is to go to the beginning. */
-	off_t ret = fr->audio_start;
+	int64_t ret = fr->audio_start;
 	*get_frame = 0;
 
 	/* But we try to find something better. */
@@ -530,39 +672,41 @@ off_t frame_fuzzy_find(mpg123_handle *fr, off_t want_frame, off_t* get_frame)
 		if(toc_entry > 99) toc_entry = 99;
 
 		/* Now estimate back what frame we get. */
-		*get_frame = (off_t) ((double)toc_entry/100. * fr->track_frames);
-		fr->accurate = FALSE;
+		*get_frame = (int64_t) ((double)toc_entry/100. * fr->track_frames);
+		fr->state_flags &= ~FRAME_ACCURATE;
 		fr->silent_resync = 1;
 		/* Question: Is the TOC for whole file size (with/without ID3) or the "real" audio data only?
 		   ID3v1 info could also matter. */
-		ret = (off_t) ((double)fr->xing_toc[toc_entry]/256.* fr->rdat.filelen);
+		ret = (int64_t) ((double)fr->xing_toc[toc_entry]/256.* fr->rdat.filelen);
 	}
 	else if(fr->mean_framesize > 0)
 	{	/* Just guess with mean framesize (may be exact with CBR files). */
 		/* Query filelen here or not? */
-		fr->accurate = FALSE; /* Fuzzy! */
+		fr->state_flags &= ~FRAME_ACCURATE; /* Fuzzy! */
 		fr->silent_resync = 1;
 		*get_frame = want_frame;
-		ret = fr->audio_start+fr->mean_framesize*want_frame;
+		ret = (int64_t) (fr->audio_start+fr->mean_framesize*want_frame);
 	}
-	debug5("fuzzy: want %li of %li, get %li at %li B of %li B",
-		(long)want_frame, (long)fr->track_frames, (long)*get_frame, (long)ret, (long)(fr->rdat.filelen-fr->audio_start));
+	debug5("fuzzy: want %" PRIi64 " of %" PRIi64
+		", get %" PRIi64 " at %" PRIi64 " B of %" PRIi64 " B"
+	,	want_frame, fr->track_frames, *get_frame, ret
+	,	(fr->rdat.filelen-fr->audio_start));
 	return ret;
 }
 
 /*
 	find the best frame in index just before the wanted one, seek to there
-	then step to just before wanted one with read_frame
+	then step to just before wanted one with INT123_read_frame
 	do not care tabout the stuff that was in buffer but not played back
 	everything that left the decoder is counted as played
 	
 	Decide if you want low latency reaction and accurate timing info or stable long-time playback with buffer!
 */
 
-off_t frame_index_find(mpg123_handle *fr, off_t want_frame, off_t* get_frame)
+int64_t INT123_frame_index_find(mpg123_handle *fr, int64_t want_frame, int64_t* get_frame)
 {
 	/* default is file start if no index position */
-	off_t gopos = 0;
+	int64_t gopos = 0;
 	*get_frame = 0;
 #ifdef FRAME_INDEX
 	/* Possibly use VBRI index, too? I'd need an example for this... */
@@ -587,7 +731,7 @@ off_t frame_index_find(mpg123_handle *fr, off_t want_frame, off_t* get_frame)
 		/* We have index position, that yields frame and byte offsets. */
 		*get_frame = fi*fr->index.step;
 		gopos = fr->index.data[fi];
-		fr->accurate = TRUE; /* When using the frame index, we are accurate. */
+		fr->state_flags |= FRAME_ACCURATE; /* When using the frame index, we are accurate. */
 	}
 	else
 	{
@@ -600,13 +744,13 @@ off_t frame_index_find(mpg123_handle *fr, off_t want_frame, off_t* get_frame)
 #ifdef FRAME_INDEX
 	}
 #endif
-	debug2("index: 0x%lx for frame %li", (unsigned long)gopos, (long) *get_frame);
+	debug2("index: 0x%" PRIx64 " for frame %" PRIi64, (uint64_t)gopos, *get_frame);
 	return gopos;
 }
 
-off_t frame_ins2outs(mpg123_handle *fr, off_t ins)
+int64_t INT123_frame_ins2outs(mpg123_handle *fr, int64_t ins)
 {	
-	off_t outs = 0;
+	int64_t outs = 0;
 	switch(fr->down_sample)
 	{
 		case 0:
@@ -617,16 +761,18 @@ off_t frame_ins2outs(mpg123_handle *fr, off_t ins)
 			outs = ins>>fr->down_sample;
 		break;
 #		ifndef NO_NTOM
-		case 3: outs = ntom_ins2outs(fr, ins); break;
+		case 3: outs = INT123_ntom_ins2outs(fr, ins); break;
 #		endif
-		default: error1("Bad down_sample (%i) ... should not be possible!!", fr->down_sample);
+		default: if(NOQUIET)
+			merror( "Bad down_sample (%i) ... should not be possible!!"
+			,	fr->down_sample );
 	}
 	return outs;
 }
 
-off_t frame_outs(mpg123_handle *fr, off_t num)
+int64_t INT123_frame_outs(mpg123_handle *fr, int64_t num)
 {
-	off_t outs = 0;
+	int64_t outs = 0;
 	switch(fr->down_sample)
 	{
 		case 0:
@@ -634,19 +780,23 @@ off_t frame_outs(mpg123_handle *fr, off_t num)
 		case 1:
 		case 2:
 #		endif
-			outs = (spf(fr)>>fr->down_sample)*num;
+			outs = (fr->spf>>fr->down_sample)*num;
 		break;
 #ifndef NO_NTOM
-		case 3: outs = ntom_frmouts(fr, num); break;
+		case 3: outs = INT123_ntom_frmouts(fr, num); break;
 #endif
-		default: error1("Bad down_sample (%i) ... should not be possible!!", fr->down_sample);
+		default: if(NOQUIET)
+			merror( "Bad down_sample (%i) ... should not be possible!!"
+			,	fr->down_sample );
 	}
 	return outs;
 }
 
-off_t frame_offset(mpg123_handle *fr, off_t outs)
+/* Compute the number of output samples we expect from this frame.
+   This is either simple spf() or a tad more elaborate for ntom. */
+int64_t INT123_frame_expect_outsamples(mpg123_handle *fr)
 {
-	off_t num = 0;
+	int64_t outs = 0;
 	switch(fr->down_sample)
 	{
 		case 0:
@@ -654,97 +804,176 @@ off_t frame_offset(mpg123_handle *fr, off_t outs)
 		case 1:
 		case 2:
 #		endif
-			num = outs/(spf(fr)>>fr->down_sample);
+			outs = fr->spf>>fr->down_sample;
 		break;
 #ifndef NO_NTOM
-		case 3: num = ntom_frameoff(fr, outs); break;
+		case 3: outs = INT123_ntom_frame_outsamples(fr); break;
 #endif
-		default: error("Bad down_sample ... should not be possible!!");
+		default: if(NOQUIET)
+			merror( "Bad down_sample (%i) ... should not be possible!!"
+			,	fr->down_sample );
+	}
+	return outs;
+}
+
+int64_t INT123_frame_offset(mpg123_handle *fr, int64_t outs)
+{
+	int64_t num = 0;
+	switch(fr->down_sample)
+	{
+		case 0:
+#		ifndef NO_DOWNSAMPLE
+		case 1:
+		case 2:
+#		endif
+			num = outs/(fr->spf>>fr->down_sample);
+		break;
+#ifndef NO_NTOM
+		case 3: num = INT123_ntom_frameoff(fr, outs); break;
+#endif
+		default: if(NOQUIET)
+			error("Bad down_sample ... should not be possible!!");
 	}
 	return num;
 }
 
 #ifdef GAPLESS
 /* input in _input_ samples */
-void frame_gapless_init(mpg123_handle *fr, off_t b, off_t e)
+void INT123_frame_gapless_init(mpg123_handle *fr, int64_t framecount, int64_t bskip, int64_t eskip)
 {
-	fr->begin_s = b;
-	fr->end_s = e;
+	debug3("INT123_frame_gapless_init: given %"PRIi64" frames, skip %"PRIi64" and %"PRIi64, framecount, bskip, eskip);
+	fr->gapless_frames = framecount;
+	if(fr->gapless_frames > 0 && bskip >=0 && eskip >= 0)
+	{
+		fr->begin_s = bskip+GAPLESS_DELAY;
+		fr->end_s = framecount*fr->spf-eskip+GAPLESS_DELAY;
+	}
+	else fr->begin_s = fr->end_s = 0;
 	/* These will get proper values later, from above plus resampling info. */
 	fr->begin_os = 0;
 	fr->end_os = 0;
-	debug2("frame_gapless_init: from %lu to %lu samples", (long unsigned)fr->begin_s, (long unsigned)fr->end_s);
+	fr->fullend_os = 0;
+	debug2("INT123_frame_gapless_init: from %"PRIi64" to %"PRIi64" samples", fr->begin_s, fr->end_s);
 }
 
-void frame_gapless_realinit(mpg123_handle *fr)
+void INT123_frame_gapless_realinit(mpg123_handle *fr)
 {
-	fr->begin_os = frame_ins2outs(fr, fr->begin_s);
-	fr->end_os   = frame_ins2outs(fr, fr->end_s);
-	debug2("frame_gapless_realinit: from %lu to %lu samples", (long unsigned)fr->begin_os, (long unsigned)fr->end_os);
+	fr->begin_os = INT123_frame_ins2outs(fr, fr->begin_s);
+	fr->end_os   = INT123_frame_ins2outs(fr, fr->end_s);
+	if(fr->gapless_frames > 0)
+	fr->fullend_os = INT123_frame_ins2outs(fr, fr->gapless_frames*fr->spf);
+	else fr->fullend_os = 0;
+
+	debug4("INT123_frame_gapless_realinit: from %"PRIi64" to %"PRIi64" samples (%"PRIi64", %"PRIi64")"
+	,	fr->begin_os, fr->end_os, fr->fullend_os, fr->gapless_frames);
 }
+
+/* At least note when there is trouble... */
+void INT123_frame_gapless_update(mpg123_handle *fr, int64_t total_samples)
+{
+	int64_t gapless_samples = fr->gapless_frames*fr->spf;
+	if(fr->gapless_frames < 1) return;
+
+	debug2("gapless update with new sample count %"PRIi64" as opposed to known %"PRIi64, total_samples, gapless_samples);
+	if(NOQUIET && total_samples != gapless_samples)
+	fprintf(stderr, "\nWarning: Real sample count %" PRIi64
+		" differs from given gapless sample count %" PRIi64 
+		". Frankenstein stream?\n", total_samples, gapless_samples);
+
+	if(gapless_samples > total_samples)
+	{
+		if(NOQUIET)
+			merror( "End sample count smaller than gapless end! (%" PRIi64
+				" < %"PRIi64"). Disabling gapless mode from now on."
+			,	total_samples, fr->end_s );
+		/* This invalidates the current position... but what should I do? */
+		INT123_frame_gapless_init(fr, -1, 0, 0);
+		INT123_frame_gapless_realinit(fr);
+		fr->lastframe = -1;
+		fr->lastoff = 0;
+	}
+}
+
 #endif
 
-/* The frame seek... This is not simply the seek to fe*spf(fr) samples in output because we think of _input_ frames here.
+/* Compute the needed frame to ignore from, for getting accurate/consistent output for intended firstframe. */
+static int64_t ignoreframe(mpg123_handle *fr)
+{
+	int64_t preshift = fr->p.preframes;
+	/* Layer 3 _really_ needs at least one frame before. */
+	if(fr->lay==3 && preshift < 1) preshift = 1;
+	/* Layer 1 & 2 reall do not need more than 2. */
+	if(fr->lay!=3 && preshift > 2) preshift = 2;
+
+	return fr->firstframe - preshift;
+}
+
+/* The frame seek... This is not simply the seek to fe*fr->spf samples in output because we think of _input_ frames here.
    Seek to frame offset 1 may be just seek to 200 samples offset in output since the beginning of first frame is delay/padding.
    Hm, is that right? OK for the padding stuff, but actually, should the decoder delay be better totally hidden or not?
    With gapless, even the whole frame position could be advanced further than requested (since Homey don't play dat). */
-void frame_set_frameseek(mpg123_handle *fr, off_t fe)
+void INT123_frame_set_frameseek(mpg123_handle *fr, int64_t fe)
 {
 	fr->firstframe = fe;
 #ifdef GAPLESS
-	if(fr->p.flags & MPG123_GAPLESS)
+	if(fr->p.flags & MPG123_GAPLESS && fr->gapless_frames > 0)
 	{
 		/* Take care of the beginning... */
-		off_t beg_f = frame_offset(fr, fr->begin_os);
+		int64_t beg_f = INT123_frame_offset(fr, fr->begin_os);
 		if(fe <= beg_f)
 		{
 			fr->firstframe = beg_f;
-			fr->firstoff   = fr->begin_os - frame_outs(fr, beg_f);
+			fr->firstoff   = fr->begin_os - INT123_frame_outs(fr, beg_f);
 		}
 		else fr->firstoff = 0;
-		/* The end is set once for a track at least, on the frame_set_frameseek called in get_next_frame() */
+		/* The end is set once for a track at least, on the INT123_frame_set_frameseek called in get_next_frame() */
 		if(fr->end_os > 0)
 		{
-			fr->lastframe  = frame_offset(fr,fr->end_os);
-			fr->lastoff    = fr->end_os - frame_outs(fr, fr->lastframe);
-		} else fr->lastoff = 0;
+			fr->lastframe  = INT123_frame_offset(fr,fr->end_os);
+			fr->lastoff    = fr->end_os - INT123_frame_outs(fr, fr->lastframe);
+		} else {fr->lastframe = -1; fr->lastoff = 0; }
 	} else { fr->firstoff = fr->lastoff = 0; fr->lastframe = -1; }
 #endif
-	fr->ignoreframe = fr->lay == 3 ? fr->firstframe-IGNORESHIFT : fr->firstframe;
+	fr->ignoreframe = ignoreframe(fr);
 #ifdef GAPLESS
-	debug5("frame_set_frameseek: begin at %li frames and %li samples, end at %li and %li; ignore from %li",
-	       (long) fr->firstframe, (long) fr->firstoff,
-	       (long) fr->lastframe,  (long) fr->lastoff, (long) fr->ignoreframe);
+	debug5("INT123_frame_set_frameseek: begin at %" PRIi64 " frames and %" PRIi64
+		" samples, end at %" PRIi64 " and %" PRIi64 "; ignore from %" PRIi64,
+		fr->firstframe,  fr->firstoff
+	,	fr->lastframe,   fr->lastoff,  fr->ignoreframe);
 #else
-	debug3("frame_set_frameseek: begin at %li frames, end at %li; ignore from %li",
-	       (long) fr->firstframe, (long) fr->lastframe, (long) fr->ignoreframe);
+	debug3("INT123_frame_set_frameseek: begin at %" PRIi64 " frames, end at %" PRIi64
+		"; ignore from %" PRIi64
+	,	fr->firstframe,  fr->lastframe,  fr->ignoreframe);
 #endif
 }
 
-void frame_skip(mpg123_handle *fr)
+void INT123_frame_skip(mpg123_handle *fr)
 {
 #ifndef NO_LAYER3
-	if(fr->lay == 3) set_pointer(fr, 512);
+	if(fr->lay == 3) INT123_set_pointer(fr, 1, 512);
 #endif
 }
 
 /* Sample accurate seek prepare for decoder. */
 /* This gets unadjusted output samples and takes resampling into account */
-void frame_set_seek(mpg123_handle *fr, off_t sp)
+void INT123_frame_set_seek(mpg123_handle *fr, int64_t sp)
 {
-	fr->firstframe = frame_offset(fr, sp);
+	fr->firstframe = INT123_frame_offset(fr, sp);
+	debug1("INT123_frame_set_seek: from %" PRIi64, fr->num);
 #ifndef NO_NTOM
-	if(fr->down_sample == 3) ntom_set_ntom(fr, fr->firstframe);
+	if(fr->down_sample == 3) INT123_ntom_set_ntom(fr, fr->firstframe);
 #endif
-	fr->ignoreframe = fr->lay == 3 ? fr->firstframe-IGNORESHIFT : fr->firstframe;
+	fr->ignoreframe = ignoreframe(fr);
 #ifdef GAPLESS /* The sample offset is used for non-gapless mode, too! */
-	fr->firstoff = sp - frame_outs(fr, fr->firstframe);
-	debug5("frame_set_seek: begin at %li frames and %li samples, end at %li and %li; ignore from %li",
-	       (long) fr->firstframe, (long) fr->firstoff,
-	       (long) fr->lastframe,  (long) fr->lastoff, (long) fr->ignoreframe);
+	fr->firstoff = sp - INT123_frame_outs(fr, fr->firstframe);
+	debug5("INT123_frame_set_seek: begin at %" PRIi64 " frames and %" PRIi64
+		" samples, end at %" PRIi64 " and %" PRIi64 "; ignore from %" PRIi64,
+		fr->firstframe,  fr->firstoff
+	,	fr->lastframe,   fr->lastoff,  fr->ignoreframe);
 #else
-	debug3("frame_set_seek: begin at %li frames, end at %li; ignore from %li",
-	       (long) fr->firstframe, (long) fr->lastframe, (long) fr->ignoreframe);
+	debug3("INT123_frame_set_seek: begin at %" PRIi64 " frames, end at %" PRIi64
+		"; ignore from %" PRIi64
+	,	 fr->firstframe,  fr->lastframe,  fr->ignoreframe);
 #endif
 }
 
@@ -754,6 +983,12 @@ int attribute_align_arg mpg123_volume_change(mpg123_handle *mh, double change)
 	return mpg123_volume(mh, change + (double) mh->p.outscale);
 }
 
+int attribute_align_arg mpg123_volume_change_db(mpg123_handle *mh, double change)
+{
+	if(mh == NULL) return MPG123_ERR;
+	return mpg123_volume(mh, dbchange(mh->p.outscale, change));
+}
+
 int attribute_align_arg mpg123_volume(mpg123_handle *mh, double vol)
 {
 	if(mh == NULL) return MPG123_ERR;
@@ -761,7 +996,7 @@ int attribute_align_arg mpg123_volume(mpg123_handle *mh, double vol)
 	if(vol >= 0) mh->p.outscale = vol;
 	else mh->p.outscale = 0.;
 
-	do_rva(mh);
+	INT123_do_rva(mh);
 	return MPG123_OK;
 }
 
@@ -788,7 +1023,7 @@ static int get_rva(mpg123_handle *fr, double *peak, double *gain)
 }
 
 /* adjust the volume, taking both fr->outscale and rva values into account */
-void do_rva(mpg123_handle *fr)
+void INT123_do_rva(mpg123_handle *fr)
 {
 	double peak = 0;
 	double gain = 0;
@@ -814,7 +1049,7 @@ void do_rva(mpg123_handle *fr)
 		debug3("changing scale value from %f to %f (peak estimated to %f)", fr->lastscale != -1 ? fr->lastscale : fr->p.outscale, newscale, (double) (newscale*peak));
 		fr->lastscale = newscale;
 		/* It may be too early, actually. */
-		if(fr->make_decode_tables != NULL) fr->make_decode_tables(fr); /* the actual work */
+		if(fr->INT123_make_decode_tables != NULL) fr->INT123_make_decode_tables(fr); /* the actual work */
 	}
 }
 
@@ -828,3 +1063,9 @@ int attribute_align_arg mpg123_getvolume(mpg123_handle *mh, double *base, double
 	return MPG123_OK;
 }
 
+int64_t attribute_align_arg mpg123_framepos64(mpg123_handle *mh)
+{
+	if(mh == NULL) return MPG123_ERR;
+
+	return mh->input_offset;
+}
